import os


def check_results_dir_exists(results_dir: str) -> bool:
    """
    Check if the results directory exists.

    :param results_dir: The path to the results directory.
    :type results_dir: str
    :return: True if the results directory exists, False otherwise.
    :rtype: bool
    """
    return os.path.exists(results_dir)



__all__ = ["WandbTracker", "get_tracker_class", "ImagePredictionLogger"]


from .callbacks import ImagePredictionLogger
from .wandb import WandbTracker


def get_tracker_class(tracker: str) -> WandbTracker:
    """
    Get tracker class by name.

    :param tracker: The name of the tracker.
    :type tracker: str
    :return: The tracker class.
    :rtype: WandbTracker
    :raises ValueError: If the tracker is not found.
    """
    if tracker == "wandb":
        return WandbTracker
    else:
        raise ValueError(f"Tracker {tracker} not found")



import argparse

from pipeline import Pipeline

parser = argparse.ArgumentParser(description="Run medicraft pipeline.")

parser.add_argument("--file", "-f", type=str, help="Path to the configuration file.", default="config.yml")
parser.add_argument("--verbose", "-v", action="store_true", help="Print verbose output.")


def main():
    """
    Main function to run the medicraft pipeline.

    This function parses the command line arguments, loads the configuration file,
    and runs the pipeline with the specified options.

    Args:
        None

    Returns:
        None
    """
    args = parser.parse_args()
    print(args.file, args.verbose)

    pipeline = Pipeline()

    pipeline.load_config(args.file)
    # pipeline.run(verbose=args.verbose)


if __name__ == "__main__":
    main()



"""
This module contains the following models:

- `GaussianDiffusion`: A model for Gaussian diffusion.
- `ResNetClassifier`: A model for ResNet classification.
"""

__all__ = ["GaussianDiffusion", "ResNetClassifier"]

from .classifier import ResNetClassifier
from .gausian_diffusion import GaussianDiffusion



import logging
import re
from datetime import datetime
from enum import Enum
from pathlib import Path

import config as cfg
import lightning as pl
import pandas as pd
import pipeline.blocks as pipeline_blocks
import torch
import torch.nn as nn
import wandb
from datasets import EyeScans, OpthalAnonymizedDataset, get_csv_dataset
from denoising_diffusion_pytorch import Unet
from generate_samples import generate_samples as generate
from lightning.pytorch.callbacks import EarlyStopping, TQDMProgressBar
from lightning.pytorch.loggers import WandbLogger
from models import GaussianDiffusion, ResNetClassifier
from pipeline.parser import parse_config, read_config_file
from torchvision import transforms as T
from trackers import ImagePredictionLogger
from trainers import Trainer
from utils import copy_results_directory
from utils.transforms import HorizontalCenterCrop


class PipelineBlocks(Enum):
    """
    Enum for the pipeline blocks

    :cvar general: Represents the general block
    :vartype general: str
    :cvar data: Represents the data block
    :vartype data: str
    :cvar experiment: Represents the experiment block
    :vartype experiment: str
    :cvar output: Represents the output block
    :vartype output: str
    """

    general = "general"
    data = "data"
    experiment = "experiment"
    output = "output"


class Pipeline:
    """
    A class to represent a pipeline
    """

    config: dict
    images_directory: str
    __df: pd.DataFrame
    __image_size: list[int]

    runned_steps: int = 0
    train_dataset: torch.utils.data.Dataset
    val_dataset: torch.utils.data.Dataset
    test_dataset: torch.utils.data.Dataset

    def load_data(self, config: pipeline_blocks.DataDTO) -> None:
        logging.debug(f"{config=}")
        self.images_directory = config.images_directory

        self.__df = get_csv_dataset(
            filepath=config.csv_file_path,
            val_size=config.validation_split,
            seed=config.split_seed,
        )

    def train_generator(
        self, config: pipeline_blocks.TrainGeneratorDTO, models_config: dict, image_size: list[int]
    ) -> None:
        """
        Train the generator model

        :param config: Configuration for training the generator model
        :type config: pipeline_blocks.TrainGeneratorDTO
        :param models_config: Configuration for the generator and diffusion models
        :type models_config: dict
        :param image_size: Size of the input images
        :type image_size: list[int]
        """
        unet_config = models_config.unet
        diffusion_config = models_config.diffusion

        diffusion = self.__get_diffusion_model(image_size, diffusion_config, unet_config)

        logging.debug(f"{config.dataset_split_type=}")
        logging.debug(f"{config.batch_size=}")
        dataset = OpthalAnonymizedDataset(
            diagnosis=config.diagnosis,
            df=self.__df[config.dataset_split_type],
            images_dir=self.images_directory,
            transform=self.transform,
            convert_image_to="L",
        )

        if config.experiment_id:
            results_folder = Path(config.results_dir) / config.experiment_id / config.diagnosis
        else:
            results_folder = Path(config.results_dir) / config.diagnosis

        trainer = Trainer(  # noqa : F841
            diffusion_model=diffusion,
            folder=self.images_directory,
            dataset=dataset,
            train_batch_size=config.batch_size,
            train_lr=config.lr,
            save_and_sample_every=config.save_and_sample_every,
            # save_and_sample_every=10,
            results_folder=results_folder,
            train_num_steps=config.num_steps,
            gradient_accumulate_every=config.gradient_accumulate_every,  # gradient accumulation steps
            ema_decay=0.995,  # exponential moving average decay
            amp=True,  # turn on mixed precision
            num_samples=9,  # number of samples to save
            calculate_fid=False,  # calculate FID during sampling
            tracker="wandb",
            tracker_kwargs={
                "tags": [config.diagnosis, "opthal_anonymized"],
                "project_name": cfg.WANDB_PRJ_NAME_TRAIN_GENERATOR,
            },
        )
        if config.start_from_checkpoint:
            trainer.load(config.start_from_checkpoint)
            logging.info(f"Model loaded from {config.start_from_checkpoint}.")
        trainer.train()
        logging.info("Training completed successfully.")
        self.runned_steps += config.num_steps

        if config.copy_results_to:
            logging.info("Copying results...")
            copy_results_directory(
                results_folder,
                Path(config.copy_results_to) / config.experiment_id if config.experiment_id else config.copy_results_to,
            )
            logging.info("Results copied successfully.")

    def train(self, config) -> None:
        """
        Train the pipeline with specified configuration
        """
        loop_blocks: list = config.loop

        only_once_blocks = [block for block in loop_blocks if not block.repeat]

        total_steps = config.total_steps

        models_config = config.models
        image_size = config.image_size

        if total_steps == 0 and self.runned_steps == 0:
            logging.info("Running 1 iteration of the pipeline")
            self.runned_steps = -1
        while total_steps > self.runned_steps:
            print(f"Step {self.runned_steps}")
            for block in loop_blocks:
                if block not in only_once_blocks and not block.repeat:
                    continue
                if block.name.lower() == pipeline_blocks.TRAIN_GENERATOR:
                    self.train_generator(block, models_config, image_size)
                elif block.name.lower() == pipeline_blocks.GENERATE_SAMPLES:
                    self.generate_samples(block, models_config, image_size)
                elif block.name.lower() == pipeline_blocks.VALIDATE:
                    self.validate(block, models_config)
                elif block.name.lower() == pipeline_blocks.FOO:
                    print("foo")
                    self.foo(block)
                if not block.repeat:
                    only_once_blocks.remove(block)
            if total_steps == 0:
                break

    def foo(self, config: pipeline_blocks.FooDTO):
        """
        Foo
        """
        print(f"{config.foo=}")
        self.runned_steps += 10

    def generate_samples(self, config: pipeline_blocks.GenerateSamplesDTO, models_config: dict, image_size: list[int]):
        """
        Generate samples
        """
        logging.info("Generating samples")

        unet_config = models_config.unet
        diffusion_config = models_config.diffusion

        diffusion = self.__get_diffusion_model(image_size, diffusion_config, unet_config)
        print("Diffusion loaded")

        if config.wandb:
            wandb.init(
                project=cfg.WANDB_PRJ_NAME_GENERATE_SAMPLES,
                tags=["opthal_anonymized", "generate_dataset"],
            )

        diffusion.load_state_dict(torch.load(config.checkpoint_path)["model"])
        # checkpoint = torch.load(config.checkpoint_path)
        logging.info(f"Model loaded from {config.checkpoint_path}.")

        generate(
            diffusion_model=diffusion,
            results_dir=config.generete_samples_dir,
            num_samples=config.num_samples,
            batch_size=config.batch_size,
        )

        if config.copy_results_to:
            logging.info("Copying results...")
            copy_results_directory(
                config.generete_samples_dir,
                str(Path(config.copy_results_to) / config.relative_dataset_results_dir / config.base_on),
            )
            logging.info("Results copied successfully.")

        raise NotImplementedError("Generating samples")

    def validate(self, config: pipeline_blocks.ValidateDTO, models_config: dict):
        """
        Validate experiment block

        :param config: The configuration for validation
        :type config: pipeline_blocks.ValidateDTO
        :param models_config: The configuration for the models
        :type models_config: dict
        """
        if config.classification:
            self.__run_classification_experiment(config.classification, models_config)

    def run(self, verbose: bool = False):
        """
        Run the pipeline.

        :param verbose: Whether to enable verbose logging (default: False)
        :type verbose: bool
        :raises ValueError: If configuration is not loaded
        """
        if verbose:
            self.__set_logging_level(
                level=20,
                save_to_file=False,
            )

        if self.config is None:
            raise ValueError("Configuration not loaded")

        # load data
        logging.info("Loading data...")
        self.load_data(self.config.get(PipelineBlocks.data.name))
        logging.info("Data loaded successfully.")

        logging.info("Training...")
        self.train(self.config.get(PipelineBlocks.experiment.name))

    def __run_classification_experiment(self, config: pipeline_blocks.ClassificationDTO, models_config: dict) -> None:
        """
        Run the classification experiment.

        Args:
            config (pipeline_blocks.ClassificationDTO): The configuration for the classification experiment.
            models_config (dict): The configuration for the models.

        Returns:
            None
        """
        classifier_config = models_config.classifier
        print(f"{classifier_config=}")
        match config.train_data_type:
            case "real":
                is_real_train_data = True
            case "synthetic":
                is_real_train_data = False
            case _:
                raise ValueError(f"Unknown data type: {config.train_data_type}")

        train_dataset_dir = config.train_dataset_dir
        val_dataset_dir = config.val_dataset_dir
        test_dataset_dir = config.test_dataset_dir

        data_module = EyeScans(
            num_workers=config.num_workers,
            batch_size=config.batch_size,
            ratio=config.ratio,
            real_word_data=is_real_train_data,
            train_data_dir=train_dataset_dir,
            val_data_dir=val_dataset_dir,
            test_dataset_dir=test_dataset_dir,
        )

        data_module.setup()
        logging.info("Data module setup completed successfully.")
        model = self.__get_classifier_model(config, classifier_config)
        wandb_logger = WandbLogger(
            project=cfg.WANDB_PRJ_NAME_CLASSIFICATION,
            id=config.logger_experiment_name,
            offline=config.offline,
            save_dir=Path(config.results_dir) / "classification-wandb",
            job_type="train",
            tags=config.logger_tags,
        )

        early_stop_callback = EarlyStopping(monitor="val_loss")
        progressbar_callback = TQDMProgressBar()

        val_samples = next(iter(data_module.val_dataloader()))

        trainer = pl.Trainer(
            max_epochs=config.epochs,
            logger=wandb_logger,
            callbacks=[early_stop_callback, ImagePredictionLogger(val_samples), progressbar_callback],
            enable_checkpointing=True,
            enable_progress_bar=True,
            log_every_n_steps=config.log_every_n_steps,
        )
        trainer.fit(model, data_module)
        trainer.test(model, data_module)

        wandb.finish()
        logging.info("Classification process completed successfully.")

    def load_config(self, config_file: str | Path = "config.yml") -> None:
        """
        Load the configuration.

        :param config_file: The path to the configuration file. Default is "config.yml".
        :type config_file: str or Path
        :return: None
        """
        config = read_config_file(config_file)
        self.config = parse_config(config)
        self.__image_size = self.config.get(PipelineBlocks.general.name).image_size
        logging.info("Configuration parsed successfully.")

    @property
    def transform(self) -> T.Compose:
        return T.Compose(
            [
                HorizontalCenterCrop(512),
                T.Resize(self.__image_size),
                T.RandomHorizontalFlip(),
                T.Grayscale(num_output_channels=1),
                T.ToTensor(),
            ]
        )

    def __get_unet_model(self, dim: int, dim_mults: list[int], channels: int) -> nn.Module:
        """
        Create a Unet model

        :param dim: The dimension of the model
        :type dim: int
        :param dim_mults: The dimension multipliers for each layer in the model
        :type dim_mults: list[int]
        :param channels: The number of input channels
        :type channels: int
        :return: The Unet model
        :rtype: nn.Module
        """
        model = Unet(
            dim=dim,
            dim_mults=dim_mults,
            channels=channels,
        )
        model.to(device=cfg.DEVICE)
        return model

    def __get_diffusion_model(self, image_size: list[int], diffusion_config: dict, unet_config: dict) -> nn.Module:
        """
        Create a Gaussian diffusion model

        :param image_size: The size of the input image
        :type image_size: list[int]
        :param diffusion_config: Configuration parameters for Gaussian diffusion
        :type diffusion_config: dict
        :param unet_config: Configuration parameters for the U-Net model
        :type unet_config: dict
        :return: The diffusion model
        :rtype: nn.Module
        """
        model = self.__get_unet_model(**unet_config)
        diffusion = GaussianDiffusion(
            model,
            image_size=image_size,
            **diffusion_config
            # loss_type = 'l1'    # L1 or L2
        )
        diffusion.to(device=cfg.DEVICE)
        logging.info(f"Model loaded to {cfg.DEVICE} device.")
        return diffusion

    def __get_classifier_model(
        self, config: pipeline_blocks.ClassificationDTO, classifier_config: dict
    ) -> pl.LightningModule:
        """
        Get the classifier model based on the provided configuration.

        Args:
            config (pipeline_blocks.ClassificationDTO): The configuration for the classifier.
            classifier_config (dict): The configuration for the specific classifier.

        Returns:
            pl.LightningModule: The classifier model.

        Raises:
            ValueError: If the architecture or loss function is unknown.
        """

        def get_loss_fn(loss_fn_name: str) -> nn.Module:
            """
            Get the loss function based on the provided name.

            Args:
                loss_fn_name (str): The name of the loss function.

            Returns:
                nn.Module: The loss function module.

            Raises:
                ValueError: If the loss function name is unknown.
            """
            if loss_fn_name.lower() == "cross_entropy":
                return nn.CrossEntropyLoss()
            elif loss_fn_name.lower() == "nll":
                return nn.NLLLoss()
            else:
                raise ValueError(f"Unknown loss function: {loss_fn_name}")

        architecture: str = classifier_config.get("architecture")
        if re.match(r"resnet", architecture.lower()):
            model = ResNetClassifier(
                architecture=architecture,
                num_classes=config.num_classes,
                loss_fn=get_loss_fn(config.loss_fn),
                pretrained=False,
                learning_rate=config.lr,
                loss_multiply=config.loss_multiply,
            )
        else:
            raise ValueError(f"Unknown architecture: {architecture}")

        return model

    def __set_logging_level(
        self,
        level: int = 10,
        save_to_file: bool = False,
        filename: str = "run.log",
    ) -> None:
        """
        Set the logging level

        :param level: The logging level to set. Valid values are:
                      10: DEBUG
                      20: INFO
                      30: WARNING
                      40: ERROR
                      50: CRITICAL
        :type level: int
        :param save_to_file: Whether to save logs to a file. Defaults to False.
        :type save_to_file: bool
        :param filename: The name of the log file. Defaults to "run.log".
        :type filename: str
        :return: None
        """
        filename = Path(filename)

        current_date = datetime.now().strftime("%Y-%m-%d-%H:%M:%S")
        filename = filename.parent / f"{current_date}_{filename.name}"

        handlers = [logging.StreamHandler()]
        if save_to_file:
            handlers.append(logging.FileHandler(filename))

        logging.basicConfig(
            level=level,
            format="%(asctime)s - %(levelname)s - %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
            force=True,
            handlers=handlers,
        )



import torch
from denoising_diffusion_pytorch import GaussianDiffusion, Trainer, Unet

DEVICE = "cuda" if torch.cuda.is_available() else "cpu"

model = Unet(
    dim=64,
    dim_mults=(1, 2, 4, 8),
    channels=1,
)
model.to(device=DEVICE)

diffusion = GaussianDiffusion(
    model,
    image_size=(512, 256)[::-1],
    # timesteps=1000,  # number of steps
    timesteps=2,  # number of steps
    # loss_type = 'l1'    # L1 or L2
)
diffusion.to(device=DEVICE)

print(f"Model loaded to {DEVICE} device.")
# data_path = "./data/healthy_eyes_1024x512"
data_path = "./data/healthy_eyes_256x128"

# data file path
trainer = Trainer(
    diffusion,
    data_path,
    train_batch_size=4,
    train_lr=2e-4,
    save_and_sample_every=2000,
    train_num_steps=200_000,  # total training steps
    gradient_accumulate_every=2,  # gradient accumulation steps
    ema_decay=0.995,  # exponential moving average decay
    amp=True,  # turn on mixed precision
    num_samples=9,  # number of samples to save
    calculate_fid=False,  # calculate FID during sampling
)
# trainer.load("healthy_eyes_512x256_s80000/model-40.pt")
trainer.train()



from functools import partial
from pathlib import Path
from typing import Literal

import pandas as pd
from denoising_diffusion_pytorch.denoising_diffusion_pytorch import convert_image_to_fn, exists
from PIL import Image
from sklearn.model_selection import train_test_split
from torch import nn
from torch.utils.data import Dataset
from torchvision import transforms as T


def get_csv_dataset(
    filepath: str | Path,
    val_size: float | None = 0.1,
    seed: int = 42,
) -> dict[str, pd.DataFrame]:
    """
    Load dataset from csv file and split it into train validation,test sets from previously splitted train/test dataset.

    :param filepath: The path to the CSV file containing the dataset.
    :type filepath: str or Path
    :param val_size: The proportion of the dataset to be used for validation. Default is 0.1 (10%).
    :type val_size: float or None
    :param seed: The random seed for reproducible train/validation split. Default is 42.
    :type seed: int
    :return: A dictionary containing the train, validation, and test sets as pandas DataFrames.
             If val_size is None, only the train and test sets are returned.
    :rtype: dict[str, pd.DataFrame]
    """
    filepath = Path(filepath) if isinstance(filepath, str) else filepath

    df = pd.read_csv(filepath)
    df = df[df["image_type"] == "OCT"]

    result = {}
    if val_size is not None:
        train_df, val_df = train_test_split(df, test_size=val_size, random_state=seed)
        result["val"] = val_df
    else:
        train_df = df[df["split"] == "train"]

    result["train"] = train_df
    result["test"] = df[df["split"] == "test"]

    return result


class OpthalAnonymizedDataset(Dataset):
    """
    Dataset class for OpthalAnonymizedDataset.

    Args:
        diagnosis (Literal["precancerous", "fluid", "benign", "reference"]): The diagnosis category for the dataset.
        df (pd.DataFrame): The DataFrame containing the dataset information.
        images_dir (str | Path): The directory path where the images are stored.
        image_size (tuple[int, int], optional): The desired size of the images. Defaults to (256, 512).
        transform (nn.Module, optional): The transformation to apply to the images. Defaults to None.
        convert_image_to (Any, optional): The function or transformation to convert the images to a specific format. Defaults to None.
        seed (int, optional): The random seed for shuffling the dataset. Defaults to None.
    """

    def __init__(
        self,
        diagnosis: Literal["precancerous", "fluid", "benign", "reference"],
        df: pd.DataFrame,
        images_dir: str | Path,
        image_size: tuple[int, int] = (256, 512),
        transform: nn.Module = None,
        convert_image_to=None,
        seed: int = None,
    ):
        self.df = df
        self.images_dir = Path(images_dir) if isinstance(images_dir, str) else images_dir
        self.df["image_path"] = self.df["filename"].apply(lambda x: self.images_dir / x)

        self.diagnosis = diagnosis
        self.diagnosis_df = self.__get_df_by_diagnosis(self.diagnosis)
        self.classes = [*self.df["diagnosis"].unique(), "reference"]

        maybe_convert_fn = (
            partial(convert_image_to_fn, convert_image_to) if exists(convert_image_to) else nn.Identity()
        )  # TODO simplify or rename

        self.transform = (
            T.Compose(
                [
                    T.Lambda(maybe_convert_fn),
                    T.CenterCrop(image_size),
                    T.Resize(image_size),
                    T.Grayscale(num_output_channels=1),
                    T.ToTensor(),
                ]
            )
            if transform is None
            else transform
        )

    def __get_df_by_diagnosis(self, diagnosis: Literal["precancerous", "fluid", "benign", "reference"]):
        match diagnosis:
            case "precancerous":
                return self.df[self.df["reference_eye"] == False][self.df["diagnosis"] == "precancerous"]  # noqa: E712
            case "fluid":
                return self.df[self.df["reference_eye"] == False][self.df["diagnosis"] == "fluid"]  # noqa: E712
            case "benign":
                return self.df[self.df["reference_eye"] == False][self.df["diagnosis"] == "benign"]  # noqa: E712
            case "reference":
                return self.df[self.df["reference_eye"] == True]  # noqa: E712

    def __getitem__(self, idx: int):
        image_path = self.diagnosis_df.iloc[idx]["image_path"]
        image = Image.open(image_path)
        return self.transform(image)

    def __len__(self):
        return self.diagnosis_df.shape[0]



from typing import Literal

import lightning as pl
import torch
import torch.nn.functional as F
import torchvision.models as models
import wandb
from torchmetrics import Accuracy, F1Score, Precision, Recall


class ResNetClassifier(pl.LightningModule):
    def __init__(
        self,
        num_classes: int,
        loss_fn=None,
        optimizer=None,
        pretrained: bool = True,
        architecture: Literal["resnet18", "resnet34", "resnet50"] = "resnet34",
        learning_rate: float = 1e-4,
        loss_multiply: float = 1.0,
        class_names: list[str] = ["fluid", "benign", "precancerous", "reference"],
    ) -> None:
        """
        Initializes the Classifier model used in validation process.

        Args:
            num_classes (int): The number of classes for classification.
            loss_fn (callable, optional): The loss function to use. Defaults to None.
            optimizer (callable, optional): The optimizer to use. Defaults to None.
            pretrained (bool, optional): Whether to use a pretrained model. Defaults to True.
            architecture (Literal["resnet18", "resnet34", "resnet50"], optional): The architecture of the model. Defaults to "resnet34".
            learning_rate (float, optional): The learning rate for the optimizer. Defaults to 1e-4.
            loss_multiply (float, optional): The loss multiplier. Defaults to 1.0.
            class_names (list[str], optional): The names of the classes. Defaults to ["fluid", "benign", "precancerous", "reference"].
        """

        self.class_names = class_names
        super().__init__()
        self.model = self.__get_model(architecture, pretrained)
        fc_output = self.model.fc.out_features

        self.head = torch.nn.Sequential(
            torch.nn.ReLU(),
            torch.nn.Linear(fc_output, num_classes),
        )

        self.save_hyperparameters()
        self.learning_rate = learning_rate

        self.loss_fn = loss_fn if loss_fn else F.nll_loss
        self.optimizer = optimizer if optimizer else torch.optim.Adam

        self.accuracy = Accuracy(task="multiclass", num_classes=num_classes, average="macro")
        self.f1score = F1Score(task="multiclass", num_classes=num_classes, average="macro")
        self.precision = Precision(task="multiclass", num_classes=num_classes, average="macro")
        self.recall = Recall(task="multiclass", num_classes=num_classes, average="macro")

        self.loss_multiply = loss_multiply

        self.validation_step_outputs = []
        self.test_step_outputs = []

    def forward(self, x):
        x = self.model(x)
        x = self.head(x)
        return x

    def training_step(self, batch, batch_idx):
        x, y = batch
        logits = self(x)
        loss = self.loss_fn(logits, y) * self.loss_multiply

        # training metrics
        preds = torch.argmax(logits, dim=1)
        acc = self.accuracy(preds, y)
        f1 = self.f1score(preds, y)
        precision = self.precision(preds, y)
        recall = self.recall(preds, y)

        self.log("train_loss", loss, on_step=True, on_epoch=True, logger=True)
        self.log("train_acc", acc, on_step=True, on_epoch=True, logger=True)
        self.log("train_f1", f1, on_step=True, on_epoch=True, logger=True)
        self.log("train_precision", precision, on_step=True, on_epoch=True, logger=True)
        self.log("train_recall", recall, on_step=True, on_epoch=True, logger=True)

        return loss

    def validation_step(self, batch, batch_idx):
        x, y = batch
        logits = self(x)
        loss = self.loss_fn(logits, y) * self.loss_multiply

        preds = torch.argmax(logits, dim=1)
        acc = self.accuracy(preds, y)
        f1 = self.f1score(preds, y)
        precision = self.precision(preds, y)
        recall = self.recall(preds, y)

        self.log("val_loss", loss, on_epoch=True, prog_bar=True)
        self.log("val_acc", acc, on_epoch=True, prog_bar=True)
        self.log("val_f1", f1, on_epoch=True, prog_bar=True)
        self.log("val_precision", precision, on_epoch=True, prog_bar=True)
        self.log("val_recall", recall, on_epoch=True, prog_bar=True)

        self.validation_step_outputs.append({"y": y.cpu(), "y_hat": preds.cpu()})
        return loss

    def on_validation_epoch_end(self):
        outputs = self.validation_step_outputs
        y_true = torch.cat([output["y"] for output in outputs]).tolist()
        y_pred = torch.cat([output["y_hat"] for output in outputs]).tolist()

        wandb.log(
            {
                "confusion_mat": wandb.plot.confusion_matrix(
                    probs=None, y_true=y_true, preds=y_pred, class_names=self.class_names
                )
            }
        )
        self.validation_step_outputs.clear()

    def test_step(self, batch, batch_idx):
        x, y = batch
        logits = self(x)
        loss = self.loss_fn(logits, y) * self.loss_multiply

        # validation metrics
        preds = torch.argmax(logits, dim=1)
        acc = self.accuracy(preds, y)
        f1 = self.f1score(preds, y)
        precision = self.precision(preds, y)
        recall = self.recall(preds, y)
        self.log("test_loss", loss, prog_bar=True)
        self.log("test_acc", acc, prog_bar=True)
        self.log("test_f1", f1, prog_bar=True)
        self.log("test_precision", precision, prog_bar=True)
        self.log("test_recall", recall, prog_bar=True)

        self.test_step_outputs.append({"y": y.cpu(), "y_hat": preds.cpu()})
        return loss

    def on_test_epoch_end(self):
        outputs = self.test_step_outputs
        y_true = torch.cat([output["y"] for output in outputs]).tolist()
        y_pred = torch.cat([output["y_hat"] for output in outputs]).tolist()

        wandb.log(
            {
                "confusion_mat": wandb.plot.confusion_matrix(
                    probs=None, y_true=y_true, preds=y_pred, class_names=self.class_names
                )
            }
        )
        self.test_step_outputs.clear()

    def configure_optimizers(self):
        optimizer = self.optimizer(self.parameters(), lr=self.learning_rate)
        return optimizer

    def __get_model(self, architecture: str, pretrained: bool) -> models.ResNet:
        match architecture:
            case "resnet18":
                model = models.resnet18(pretrained=pretrained)
            case "resnet34":
                model = models.resnet34(pretrained=pretrained)
            case "resnet50":
                model = models.resnet50(pretrained=pretrained)
            case _:
                raise ValueError(f"Invalid architecture: {architecture}")
        return model



import logging
from pathlib import Path
from shutil import copyfile

from tqdm import tqdm


def copy_results_directory(src_dir: str, dest_dir: str, absolute_paths: tuple[bool] = (False, False)) -> None:
    """
    Copy the results directory to the destination directory.

    :param src_dir: Source directory.
    :param dest_dir: Destination directory.
    :param absolute_paths: Tuple of two booleans, if True, the source or destination paths will be absolute.
    :return: None
    :raises FileNotFoundError: If the source directory does not exist.
    :raises NotADirectoryError: If the source directory is not a directory.
    """
    src_dir = Path(src_dir) if not absolute_paths[0] else Path("/") / src_dir
    dest_dir = Path(dest_dir) if not absolute_paths[1] else Path("/") / dest_dir

    if not src_dir.exists():
        raise FileNotFoundError(f"Source directory {src_dir} does not exist.")
    if not src_dir.is_dir():
        raise NotADirectoryError(f"Source directory {src_dir} is not a directory.")

    if not dest_dir.exists():
        logging.info(f"Creating destination directory {dest_dir}")
        dest_dir.mkdir(parents=True)

    total_num_files = sum(1 for _ in src_dir.rglob("*") if _.is_file())

    for src_file in tqdm(src_dir.rglob("**/*"), total=total_num_files, desc="Coping files"):
        if src_file.is_file():
            relative_path = src_file.relative_to(src_dir)
            dest_file = dest_dir / relative_path
            dest_file.parent.mkdir(parents=True, exist_ok=True)
            tqdm.write(f"Copying {src_file} to {dest_file}")
            copyfile(src_file, dest_file)



import logging
from pathlib import Path

import torch
import wandb
from denoising_diffusion_pytorch import Unet
from denoising_diffusion_pytorch.denoising_diffusion_pytorch import num_to_groups
from ema_pytorch import EMA
from PIL import Image
from torchvision import utils
from tqdm import tqdm

from models import GaussianDiffusion
from utils import copy_results_directory

DEVICE = "cuda" if torch.cuda.is_available() else "cpu"


model = Unet(dim=64, dim_mults=(1, 2, 4, 8), channels=1)
model.to(device=DEVICE)

diffusion = GaussianDiffusion(
    model,
    image_size=(512, 256)[::-1],
    timesteps=1000,  # number of steps
    # timesteps=2,  # number of steps
    # loss_type = 'l1'    # L1 or L2
)
diffusion.to(device=DEVICE)


def generate_samples(
    diffusion_model: GaussianDiffusion,
    results_dir: str,
    # ema_model=,
    ema_decay=0.995,
    ema_update_every=10,
    num_samples: int = 100,
    batch_size: int = 1,
    start_sample_idx: int = 0,
):
    # noise = torch.randn([16, 1, 256, 512], device=DEVICE)
    if not Path(results_dir).exists():
        Path(results_dir).mkdir(parents=True)

    ema = EMA(
        diffusion_model,
        beta=ema_decay,
        update_every=ema_update_every,
    )
    ema.to(DEVICE)

    with torch.inference_mode():
        batches = num_to_groups(
            num_samples,
            batch_size,
        )
        # print(batches)
        t = tqdm(total=num_samples)
        for idx, batch_size in enumerate(batches):
            # batch = batches[idx]
            images: list[torch.Tensor] = ema.ema_model.sample(batch_size=batch_size)
            for i in range(len(images)):
                # detach image from tensor to pil image

                # numpy_array = images[i].add_(0.5).clamp_(0, 255).permute(1, 2, 0).to("cpu", torch.uint8).numpy()
                # pil_image = Image.fromarray(numpy_array)
                progress = i + idx * batch_size
                utils.save_image(images[i], f"{results_dir}/image_{(start_sample_idx + progress):>06}.png")
                # save_image(pil_image, f"{results_dir}/image_{progress}.png")
                t.update(progress)
    t.close()


@torch.no_grad()
def p_sample_loop(
    self,
    model,
    noise,
    device,
    noise_fn=torch.randn,
    capture_every=1000,
):
    img = noise
    imgs = []

    for i in tqdm(reversed(range(self.num_timesteps)), total=self.num_timesteps):
        img = self.p_sample(
            model,
            img,
            torch.full((img.shape[0],), i, dtype=torch.int64).to(device),
            noise_fn=noise_fn,
        )

        if i % capture_every == 0:
            imgs.append(img)

    imgs.append(img)

    return imgs


def save_image(image: Image, path) -> None:
    with open(path, "wb") as f:
        image.save(f, format="png")


if __name__ == "__main__":
    DATASET_PATH = "datasets/05.2024/0001"

    ONEDRIVE_DATASETS_DIR = "/home/wmi/OneDrive/General/results/datasets/ophtal_anonym/"
    ONEDRIVE_DESTINATION_DIR = str(Path(ONEDRIVE_DATASETS_DIR) / Path(DATASET_PATH).name)

    Path.mkdir(Path(DATASET_PATH), parents=True, exist_ok=True)
    Path.mkdir(Path(ONEDRIVE_DESTINATION_DIR), parents=True, exist_ok=True)

    print(ONEDRIVE_DATASETS_DIR)
    models = {
        "reference": ".results/reference/model-100.pt",
        "benign": ".results/benign/model-125.pt",
        "fluid": ".results/fluid/model-125.pt",
        "precancerous": ".results/precancerous/model-125.pt",
    }

    wandb.init(project="opthal_anonymized_datasets", tags=["opthal_anonymized", "generate_dataset"])

    for model_name, model_path in models.items():
        diffusion.load_state_dict(torch.load(model_path)["model"])
        checkpoint = torch.load(model_path)  # TODO unused variable ?
        dataset_path = Path(DATASET_PATH) / model_name
        logging.info(f"Model {model_name} loaded from {model_path}.")
        logging.info(f"Generating samples for {model_name}.")
        if model_name != "reference":
            generate_samples(
                diffusion,
                dataset_path,
                num_samples=1000,
                batch_size=10,
                # start_sample_idx=2010,
            )
        logging.info(f"Samples for {model_name} generated.")
        copy_results_directory(DATASET_PATH, ONEDRIVE_DESTINATION_DIR)
        logging.info(f"Samples for {model_name} copied.")



from pathlib import Path

import torch
from denoising_diffusion_pytorch.denoising_diffusion_pytorch import num_to_groups
from ema_pytorch import EMA
from torchvision import utils
from tqdm import tqdm

from config import DEVICE
from models import GaussianDiffusion


def generate_samples(
    diffusion_model: GaussianDiffusion,
    results_dir: str,
    ema_decay=0.995,
    ema_update_every=10,
    num_samples: int = 100,
    batch_size: int = 1,
    start_sample_idx: int = 0,
):
    """
    Generate samples using the given diffusion model and save them to the specified directory.

    Args:
        diffusion_model (GaussianDiffusion): The diffusion model used for generating samples.
        results_dir (str): The directory where the generated samples will be saved.
        ema_decay (float, optional): The decay rate for the exponential moving average (EMA) of the diffusion model.
            Defaults to 0.995.
        ema_update_every (int, optional): The number of steps between EMA updates. Defaults to 10.
        num_samples (int, optional): The total number of samples to generate. Defaults to 100.
        batch_size (int, optional): The batch size used for generating samples. Defaults to 1.
        start_sample_idx (int, optional): The starting index for the generated sample filenames. Defaults to 0.
    """

    if not Path(results_dir).exists():
        Path(results_dir).mkdir(parents=True)

    ema = EMA(
        diffusion_model,
        beta=ema_decay,
        update_every=ema_update_every,
    )
    ema.to(DEVICE)

    with torch.inference_mode():
        batches = num_to_groups(
            num_samples,
            batch_size,
        )
        t = tqdm(total=num_samples)
        for idx, batch_size in enumerate(batches):
            images: list[torch.Tensor] = ema.ema_model.sample(batch_size=batch_size)
            for i in range(len(images)):
                progress = i + idx * batch_size
                utils.save_image(images[i], f"{results_dir}/image_{(start_sample_idx + progress):>06}.png")
                t.update(progress)
    t.close()



__all__ = [
    "ConfigBlocks",
    "FOO",
    "GENERATE_SAMPLES",
    "TRAIN_GENERATOR",
    "VALIDATE",
    "FooDTO",
    "GenerateSamplesDTO",
    "TrainGeneratorDTO",
    "ValidateDTO",
    "DataDTO",
    "ClassificationDTO",
]

from .blocks import ConfigBlocks
from .data_dto import DataDTO
from .experiment_dto import (
    FOO,
    GENERATE_SAMPLES,
    TRAIN_GENERATOR,
    VALIDATE,
    ClassificationDTO,
    FooDTO,
    GenerateSamplesDTO,
    TrainGeneratorDTO,
    ValidateDTO,
)



from pathlib import Path
from typing import Literal, Optional, Union

from pipeline.blocks.models_dto import ModelsDTO
from pydantic import BaseModel, ConfigDict, computed_field, field_validator

TRAIN_GENERATOR = "train_generator"
GENERATE_SAMPLES = "generate_samples"
VALIDATE = "validate"
FOO = "foo"


class LoopObjectDTO(BaseModel):
    model_config = ConfigDict(protected_namespaces=())
    name: str
    repeat: bool = True

    @field_validator("name")
    def name_validator(cls, v):
        if v not in [TRAIN_GENERATOR, GENERATE_SAMPLES, VALIDATE, FOO]:
            raise ValueError("Undefined training loop object")
        return v.title()


class TrainGeneratorDTO(LoopObjectDTO):
    batch_size: int
    lr: float
    save_and_sample_every: int
    num_steps: int
    results_dir: str
    gradient_accumulate_every: int = 4

    experiment_id: Optional[str] = None
    copy_results_to: Optional[str] = None
    start_from_checkpoint: Optional[str] = None
    dataset_split_type: Literal["train", "val", "test"] = "train"
    diagnosis: Literal["precancerous", "fluid", "benign", "reference"]

    @field_validator("name")
    def name_validator(cls, v):
        if v != TRAIN_GENERATOR:
            raise ValueError("name must be 'train_generator'")
        return v.title()


class GenerateSamplesDTO(LoopObjectDTO):
    num_samples: int
    batch_size: int
    wandb: bool = True

    model_version: str
    base_on: str
    results_dir: str
    copy_results_to: Optional[str] = None
    checkpoint: Optional[str] = None
    relative_dataset_results_dir: str = "dataset"

    @computed_field
    @property
    def checkpoint_path(self) -> str:
        if self.checkpoint:
            return self.checkpoint
        return str(Path(self.results_dir) / self.base_on / f"{self.model_version}.pt")

    @computed_field
    @property
    def generete_samples_dir(self) -> str:
        return str(Path(self.results_dir) / self.relative_dataset_results_dir / self.base_on)

    @field_validator("name")
    def name_validator(cls, v):
        if v != GENERATE_SAMPLES:
            raise ValueError(f"name must be {GENERATE_SAMPLES}")
        return v.title()


class FooDTO(LoopObjectDTO):
    foo: str = "foo"

    @field_validator("name")
    def name_validator(cls, v):
        if v != FOO:
            raise ValueError(f"name must be {FOO}")
        return v.title()


class ClassificationDTO(BaseModel):
    epochs: int
    lr: float
    loss_multiply: float = 1.0
    ratio: list[float] = [0.8, 0.2]
    class_names: list[str] = ["fluid", "benign", "precancerous", "reference"]

    train_data_type: Literal["real", "synthetic"] = "synthetic"

    train_dataset_dir: str
    val_dataset_dir: Optional[str] = None
    test_dataset_dir: str

    loss_fn: str = "cross_entropy"
    num_workers: int = 4
    batch_size: int = 32

    log_every_n_steps: int = 10
    logger_tags: Optional[list[str]] = None
    logger_experiment_name: Optional[str] = None
    offline: bool = False

    results_dir: str

    @computed_field
    @property
    def num_classes(self) -> int:
        return len(self.class_names)


class ValidateDTO(LoopObjectDTO):
    results_dir: str
    copy_results_to: Optional[str] = None

    classification: Optional[ClassificationDTO] = None

    @field_validator("name")
    def name_validator(cls, v):
        if v != VALIDATE:
            raise ValueError(f"name must be {VALIDATE}")
        return v.title()

    @field_validator("classification", mode="before")
    @classmethod
    def classification_validator(cls, v, values):
        v["results_dir"] = values.data.get("results_dir", ".results")
        return ClassificationDTO(**v)


class ExperimentDTO(BaseModel):
    total_steps: int
    image_size: list[int]
    models: ModelsDTO
    results_dir: str
    copy_results_to: Optional[str] = None
    loop: list[Union[TrainGeneratorDTO, GenerateSamplesDTO, ValidateDTO, FooDTO]]

    @field_validator("loop", mode="before")
    @classmethod
    def loop_validator(cls, v, values):
        res = []
        for loop_obj in v:
            loop_obj["results_dir"] = values.data.get("results_dir", ".results")
            loop_obj["copy_results_to"] = values.data.get("copy_results_to", None)
            if loop_obj["name"] == TRAIN_GENERATOR:
                res.append(TrainGeneratorDTO(**loop_obj))
            if loop_obj["name"] == GENERATE_SAMPLES:
                res.append(GenerateSamplesDTO(**loop_obj))
            if loop_obj["name"] == VALIDATE:
                res.append(ValidateDTO(**loop_obj))
            if loop_obj["name"] == FOO:
                res.append(FooDTO(**loop_obj))
        return res



from pathlib import Path
from shutil import copy2
from typing import Optional

from PIL import Image
from tqdm import tqdm

DATA_ROOT_PATH = Path("data")
EASY_CASES_PATH = DATA_ROOT_PATH / "easy_cases_with_fluid"
REFERENCE_EYE_DIR_NAME = "healthy_eye"  # TODO rename to reference_eye

IMAGE_FORMATS = [".jpg", ".jpeg", ".png"]


def get_patients_paths(dataset_dir: Path) -> list[Path]:
    """
    Returns a list of paths to patients directories in the given directory.

    :param dataset_dir: The directory containing the dataset.
    :type dataset_dir: Path
    :return: A list of Path objects representing the paths to patients directories.
    :rtype: list[Path]
    """
    return [p for p in dataset_dir.iterdir() if p.is_dir()]


def get_images_paths(images_dir: Path) -> Optional[list[Path]]:
    """
    Returns a list of paths to images in the given directory.

    :param images_dir: The directory containing the images.
    :type images_dir: Path
    :return: A list of paths to images.
    :rtype: Optional[list[Path]]
    """
    try:
        return [p for p in images_dir.iterdir() if p.suffix.lower() in IMAGE_FORMATS]
    except FileNotFoundError:
        return None


def get_lesion_eyes_paths(dataset_dir: Path) -> Optional[list[Path]]:
    """
    Returns a list of paths to all reference (second, healthy) eyes images in the dataset directory.

    :param dataset_dir: The directory path of the dataset.
    :type dataset_dir: Path
    :return: A list of paths to the reference eyes images.
    :rtype: Optional[list[Path]]
    """
    lesion_eyes_paths = []

    for patient_dir in get_patients_paths(dataset_dir):
        patient_examinations = [p for p in patient_dir.iterdir() if p.is_dir() and p.name != REFERENCE_EYE_DIR_NAME]
        for examination_dir in patient_examinations:
            lesion_eye_photos = get_images_paths(examination_dir)
            if lesion_eye_photos:
                lesion_eyes_paths.extend(lesion_eye_photos)
    return lesion_eyes_paths


def get_reference_eyes_paths(dataset_dir: Path) -> Optional[list[Path]]:
    """
    Returns a list of paths to images of healthy eyes with reference healthy eye.

    :param dataset_dir: The directory path of the dataset.
    :type dataset_dir: Path
    :return: A list of paths to images of healthy eyes.
    :rtype: Optional[list[Path]]
    """
    healthy_eyes_paths = []

    for patient_dir in get_patients_paths(dataset_dir):
        healthy_eye_dir = patient_dir / REFERENCE_EYE_DIR_NAME
        healthy_eye_photos = get_images_paths(healthy_eye_dir)
        if healthy_eye_photos:
            healthy_eyes_paths.extend(healthy_eye_photos)
    return healthy_eyes_paths


def resize_images_and_save(
    images_paths: list[Path],
    output_dir_path: str,
    size: tuple[int, int],
    max_images: int | None = None,
) -> None:
    """
    Resizes images from given paths and saves them to the given directory.

    :param images_paths: List of paths to the images.
    :type images_paths: list[Path]
    :param output_dir_path: Path to the output directory.
    :type output_dir_path: str
    :param size: The desired size of the images after resizing.
    :type size: tuple[int, int]
    :param max_images: The maximum number of images to resize and save. If None, all images will be processed.
    :type max_images: int | None
    """
    images_paths = images_paths[:max_images] if max_images else images_paths

    output_dir_path = Path(output_dir_path)
    output_dir_path.mkdir(parents=True, exist_ok=True)

    for image_path in tqdm(images_paths):
        image = Image.open(image_path).resize(size)
        image.save(output_dir_path / image_path.name)


def copy_images_to_dir(images_paths: list[Path], destination_dir: Path) -> None:
    """
    Copies images from given paths to given directory.

    :param images_paths: List of paths to the images.
    :type images_paths: list[Path]
    :param destination_dir: Path to the destination directory.
    :type destination_dir: Path
    :return: None
    """
    destination_dir.mkdir(parents=True, exist_ok=True)
    for source_path in tqdm(images_paths):
        destination_path = destination_dir / source_path.name
        copy2(source_path, destination_path)


if __name__ == "__main__":
    patiens_paths = get_patients_paths(EASY_CASES_PATH)

    # print(get_healty_eyes_paths(patiens_paths[0]))
    healthy_eyes_paths = get_reference_eyes_paths(EASY_CASES_PATH)

    # raise ValueError("Testing only")
    print(f"Healthy eyes: {len(healthy_eyes_paths)}")
    print(healthy_eyes_paths[0])
    lesion_eyes_paths = get_lesion_eyes_paths(EASY_CASES_PATH)
    print(f"Lesion eyes : {len(lesion_eyes_paths)}")
    print(lesion_eyes_paths[0])

    # destination_dir = DATA_ROOT_PATH / "healthy_eyes"
    # copy_images_to_dir(healthy_eyes_paths, destination_dir)

    # SIZE = (256, 128)
    # SIZE = (128, 64)
    SIZE = (512, 256)
    print(f"Resize to {SIZE}")
    HEALTHY_OUTPUT_DIR = "datasets/genuine_eyes_ballanced/healthy_eyes_" + str("x".join([str(x) for x in SIZE]))
    WITH_FLUID_OUTPUT_DIR = "datasets/genuine_eyes_ballanced/with_fluid_eyes_" + str("x".join([str(x) for x in SIZE]))
    resize_images_and_save(healthy_eyes_paths, HEALTHY_OUTPUT_DIR, SIZE, max_images=250)
    resize_images_and_save(lesion_eyes_paths, WITH_FLUID_OUTPUT_DIR, SIZE, max_images=250)



import torch
from denoising_diffusion_pytorch import Unet

import config as cfg
from datasets import OpthalAnonymizedDataset
from models import GaussianDiffusion
from trainers import Trainer

DEVICE = "cuda" if torch.cuda.is_available() else "cpu"


def main():
    dataset = OpthalAnonymizedDataset(
        diagnosis="reference",
        csv_dataset_file=cfg.DATASET_FILE_PATH,
    )
    model = Unet(
        dim=64,
        dim_mults=(1, 2, 4, 8),
        channels=1,
    )

    model.to(device=DEVICE)

    diffusion = GaussianDiffusion(
        model,
        image_size=(512, 256)[::-1],
        # timesteps=1000,  # number of steps #
        timesteps=2,  # number of steps
        # loss_type = 'l1'    # L1 or L2
    )
    diffusion.to(device=DEVICE)
    print(f"Model loaded to {DEVICE} device.")

    # data_path = "./data/input_datasets/with_fluid_eyes_512x256"
    # data_path = "./data/input_datasets/healthy_eyes_512x256"

    trainer = Trainer(  # noqa : F841
        diffusion,
        str(cfg.DATASET_FILE_PATH.parent / "images"),
        dataset=dataset,
        train_batch_size=1,
        train_lr=2e-4,
        # save_and_sample_every=2000,
        save_and_sample_every=10,
        results_folder="./.results/reference",
        train_num_steps=150_000,  # total training steps
        gradient_accumulate_every=4,  # gradient accumulation steps
        ema_decay=0.995,  # exponential moving average decay
        amp=True,  # turn on mixed precision
        num_samples=9,  # number of samples to save
        calculate_fid=False,  # calculate FID during sampling
        tracker="wandb",
        tracker_kwargs={
            "tags": ["reference_eyes"],
            "mode": "offline",
        },
    )
    # trainer.load(".results/sleek-glitter-5/model-75.pt")
    trainer.train()


if __name__ == "__main__":
    main()



import torch
import wandb
from lightning import Callback


class ImagePredictionLogger(Callback):
    """
    Callback for logging image predictions during validation.

    Args:
        val_samples (tuple): A tuple containing validation images and labels.
        num_samples (int): Number of samples to log. Default is 32.
    """

    def __init__(self, val_samples: tuple, num_samples: int = 32):
        super().__init__()
        self.num_samples = num_samples
        self.val_imgs, self.val_labels = val_samples

    def on_validation_epoch_end(self, trainer, pl_module):
        """
        Method called at the end of each validation epoch.

        Args:
            trainer (Trainer): The PyTorch Lightning Trainer object.
            pl_module (LightningModule): The PyTorch Lightning module being trained.
        """
        val_imgs = self.val_imgs.to(device=pl_module.device)
        val_labels = self.val_labels.to(device=pl_module.device)
        logits = pl_module(val_imgs)
        preds = torch.argmax(logits, -1)

        # Log the images as wandb Image
        trainer.logger.experiment.log(
            {
                "examples": [
                    wandb.Image(x, caption=f"Pred:{pred}, Label:{y}")
                    for x, pred, y in zip(
                        val_imgs[: self.num_samples], preds[: self.num_samples], val_labels[: self.num_samples]
                    )
                ]
            }
        )



import argparse

import lightning as pl
import torch.nn as nn
from lightning.pytorch.callbacks import EarlyStopping, TQDMProgressBar
from lightning.pytorch.loggers import WandbLogger

import wandb
from datasets import EyeScans
from models import ResNetClassifier
from trackers import ImagePredictionLogger

parser = argparse.ArgumentParser(description="Run a classification experiment.")
parser.add_argument("--real", action="store_true", help="Use real-world data.")


def main():
    TRAIN_DATASET_DIR = "data/datasets/ophthal_anonym_classed-stratified/train"
    VAL_DATASET_DIR = "data/datasets/ophthal_anonym_classed-stratified/val"
    TEST_DATASET_DIR = "data/datasets/ophthal_anonym_classed-stratified/test"

    SYNTH_TRAIN_DATASET_DIR = "data/datasets/synthetic/0001"

    args = parser.parse_args()
    if args.real:
        REAL_WORD_DATA = True
    else:
        REAL_WORD_DATA = False

    train_data_dir, val_data_dir = (
        (TRAIN_DATASET_DIR, VAL_DATASET_DIR) if REAL_WORD_DATA else (SYNTH_TRAIN_DATASET_DIR, None)
    )

    dm = EyeScans(
        num_workers=4,
        batch_size=32,
        ratio=[0.8, 0.2],
        real_word_data=REAL_WORD_DATA,
        train_data_dir=train_data_dir,
        val_data_dir=val_data_dir,
        test_dataset_dir=TEST_DATASET_DIR,
    )
    dm.setup()

    # Samples required by the custom ImagePredictionLogger callback to log image predictions.
    val_samples = next(iter(dm.val_dataloader()))
    # val_imgs, val_labels = val_samples[0], val_samples[1]
    # print(val_imgs.shape, val_labels.shape)

    # raise

    loss_multiply = 1.0

    model = ResNetClassifier(
        num_classes=4,
        loss_fn=nn.CrossEntropyLoss(),
        pretrained=False,
        learning_rate=1e-3,
        loss_multiply=loss_multiply,
    )

    dataset_type = "real" if REAL_WORD_DATA else "synthetic"
    wandb_logger = WandbLogger(
        project="medicraft-classification",
        # mode="offline",
        job_type="train",
        tags=[
            "resnet34",
            # "pretrained",
            dataset_type,
            "center-crop",
            "average-macro",
            "trained-on-straified",
            "tested-on-straified",
        ],
    )
    early_stop_callback = EarlyStopping(monitor="val_loss")
    # checkpoint_callback = ModelCheckpoint()
    progressbar_callback = TQDMProgressBar()
    trainer = pl.Trainer(
        max_epochs=15,
        # progress_bar_refresh_rate=20,
        # gpus=1,
        logger=wandb_logger,
        callbacks=[early_stop_callback, ImagePredictionLogger(val_samples), progressbar_callback],
        # checkpoint_callback=checkpoint_callback,
        enable_checkpointing=True,
        enable_progress_bar=True,
        log_every_n_steps=10,
    )

    trainer.fit(model, dm)
    trainer.test(model, dm)

    wandb.finish()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['medicraft',
 'medicraft.datasets',
 'medicraft.experiments',
 'medicraft.models',
 'medicraft.pipeline',
 'medicraft.pipeline.blocks',
 'medicraft.trackers',
 'medicraft.trainers',
 'medicraft.utils',
 'medicraft.validation']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML==6.0.1',
 'denoising-diffusion-pytorch==1.9.4',
 'ema-pytorch==0.3.1',
 'lightning==2.2.1',
 'matplotlib==3.8.4',
 'pandas==2.2.2',
 'pillow==10.3.0',
 'pydantic==2.7.2',
 'scikit-learn==1.4.2',
 'seaborn==0.13.2',
 'torch>=2.1.1,<3.0.0',
 'torchmetrics==1.3.1',
 'torchvision>=0.16.1,<0.17.0',
 'tqdm==4.66.2',
 'umap-learn==0.5.5',
 'wandb==0.16.3']

setup_kwargs = {
    'name': 'medicraft',
    'version': '0.9.2',
    'description': 'Medicraft synthetic dataset generator',
    'long_description': '# Medicraft\n\n\n## About \n\nThe medical imaging analysis using machine learning encounters challenges such as limited and imbalanced datasets, often constrained by privacy concerns related to patient information. The Medicraft project addresses these limitations by developing software capable of generating synthetic and diverse medical datasets from imaging information. In cooperation with the University Clinical Hospital in Poznan this tool utilizes OCT eye scans, featuring images with abnormalities like tumors and melanomas. To reduce the scarcity of real data, for medically rare cases, the solution uses diffusion models (Denoising Diffusion Probabilistic Model) to create synthetic balanced datasets which can facilitate better research and education in machine learning for medical image analysis and classification. The project uses a data-driven learning approach focused on analyzing the generated data, where synthetic images are analyzed for accuracy with the medical team, in order to achieve the best possible result.\n\n\n## Documentation\nThe documentation for the project can be found [here](https://drfifonz.github.io/medicraft/index.html).\n\n## Installation\n\n###  Using conda (recomended)\n Create an environment from the `environment.yml` file:\n ```bash\n conda env create -f environment.yml\n```\n### Using pip\nInstall the required packages using pip:\n```bash\npip install -r requirements.txt\n```\n\n### As a package (torch dependencies needs to be solved manually)\nTo install the project as a package, run the following command:\n```bash\npip install medicraft\n```\n\n## Post-Installation Steps\nAfter successfully installing Medicraft, it is highly recommended to familiarize yourself with the documentation. You can find the detailed documentation [here](https://drfifonz.github.io/medicraft/index.html).\n\n## Usage\nTo learn how to run the project, refer to the usage instructions provided in the documentation [here](https://drfifonz.github.io/medicraft/usage.html).\n\nFor running a project create `config.yml` file and run the following command:\n```bash\npython src/main.py -f config.yml\n```\nSee config examples here: [configs](https://drfifonz.github.io/medicraft/examples.html#example-section)\n\n\n\n\nFor more information on how to install the project, please refer to the [installation guide](https://drfifonz.github.io/medicraft/installation.html).\n\n---\n ### The conference poster provides information about the project. You can find it [here](docs/conference_poster.pdf).\n',
    'author': 'Filip Patyk',
    'author_email': 'fp.patyk@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/drfifonz/medicraft',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.11,<4.0',
}


setup(**setup_kwargs)



from typing import Any, Optional

from pydantic import BaseModel


class ModelsDTO(BaseModel):
    unet: Optional[dict[str, Any]] = None
    diffusion: Optional[dict[str, Any]] = None
    classifier: Optional[dict[str, Any]] = None



from pathlib import Path

from pydantic import BaseModel, computed_field, field_validator


class DataDTO(BaseModel):
    csv_file_path: str
    validation_split: float = 0.2

    split_seed: int = 42

    @field_validator("validation_split", mode="before")
    @classmethod
    def validation_split_validator(cls, v):
        if isinstance(v, float) and 0 <= v < 1:
            return v
        elif isinstance(v, int) and 0 <= v < 100:
            return v / 100
        elif isinstance(v, str) and v.endswith("%"):
            return int(v[:-1]) / 100
        elif isinstance(v, str):
            return eval(v)
        else:
            raise ValueError("validation_split must be a float between 0 and 1")

    @computed_field
    @property
    def images_directory(self) -> str:
        return str(Path(self.csv_file_path).parent / "images")



import math
import os
from pathlib import Path
from typing import Literal

import torch
from denoising_diffusion_pytorch import Trainer as DiffusionTrainer
from denoising_diffusion_pytorch.denoising_diffusion_pytorch import Dataset as _Dataset
from denoising_diffusion_pytorch.denoising_diffusion_pytorch import (
    cpu_count,
    cycle,
    divisible_by,
    exists,
    num_to_groups,
)
from denoising_diffusion_pytorch.version import __version__
from torch import nn
from torch.utils.data import DataLoader, Dataset
from torchvision import utils
from tqdm.auto import tqdm

from trackers.wandb import WandbTracker


class Trainer(DiffusionTrainer):
    def __init__(
        self,
        diffusion_model: nn.Module,
        folder: str = None,
        *,
        dataset: Dataset = None,
        train_batch_size: int = 16,
        gradient_accumulate_every: int = 1,
        augment_horizontal_flip: int = True,
        train_lr: float = 1e-4,
        train_num_steps: int = 100_000,
        break_every_steps: int | None = None,
        ema_update_every: int = 10,
        ema_decay: float = 0.995,
        adam_betas: tuple[float, float] = (0.9, 0.99),
        save_and_sample_every: int = 1000,
        num_samples: int = 25,
        results_folder: str = "./results",
        amp: bool = False,
        mixed_precision_type: str = "fp16",
        split_batches: bool = True,
        convert_image_to: Literal["L", "RGB", "RGBA"] | None = None,
        calculate_fid: bool = True,
        inception_block_idx: int = 2048,
        max_grad_norm: float = 1.0,
        num_fid_samples: int = 50000,
        save_best_and_latest_only: bool = False,
        tracker: str | None = None,
        tracker_kwargs: dict | None = None,
    ):
        """
        Initialize the Trainer object.

        Args:
            diffusion_model (nn.Module): The diffusion model to be trained.
            folder (str, optional): The folder path containing the dataset. Defaults to None.
            dataset (Dataset, optional): The dataset object. Defaults to None.
            train_batch_size (int, optional): The batch size for training. Defaults to 16.
            gradient_accumulate_every (int, optional): Number of steps to accumulate gradients before performing optimization. Defaults to 1.
            augment_horizontal_flip (int, optional): Whether to apply horizontal flip augmentation. Defaults to True.
            train_lr (float, optional): The learning rate for training. Defaults to 1e-4.
            train_num_steps (int, optional): The total number of training steps. Defaults to 100_000.
            break_every_steps (int | None, optional): Number of steps after which to break the training loop. Defaults to None.
            ema_update_every (int, optional): Number of steps after which to update the exponential moving average of model weights. Defaults to 10.
            ema_decay (float, optional): Decay rate for the exponential moving average. Defaults to 0.995.
            adam_betas (tuple[float, float], optional): Coefficients for computing running averages of gradient and its square. Defaults to (0.9, 0.99).
            save_and_sample_every (int, optional): Number of steps after which to save model checkpoints and generate samples. Defaults to 1000.
            num_samples (int, optional): Number of samples to generate during training. Defaults to 25.
            results_folder (str, optional): The folder path to save results. Defaults to "./results".
            amp (bool, optional): Whether to use automatic mixed precision training. Defaults to False.
            mixed_precision_type (str, optional): The type of mixed precision training. Defaults to "fp16".
            split_batches (bool, optional): Whether to split batches across devices. Defaults to True.
            convert_image_to (Literal["L", "RGB", "RGBA"] | None, optional): The color space to convert images to. Defaults to None.
            calculate_fid (bool, optional): Whether to calculate Fréchet Inception Distance (FID) during training. Defaults to True.
            inception_block_idx (int, optional): The index of the inception block to use for FID calculation. Defaults to 2048.
            max_grad_norm (float, optional): Maximum gradient norm for gradient clipping. Defaults to 1.0.
            num_fid_samples (int, optional): Number of samples to use for FID calculation. Defaults to 50000.
            save_best_and_latest_only (bool, optional): Whether to save only the best and latest model checkpoints. Defaults to False.
            tracker (str | None, optional): The tracker name for experiment tracking. Defaults to None.
            tracker_kwargs (dict | None, optional): Additional keyword arguments for the tracker. Defaults to None.
        """

        self.tracker = None  # TODO try if is it necessary then remove this line
        if tracker:
            parameters = {k: v for k, v in locals().items() if k != "self"}
            [
                parameters.pop(key)
                for key in [
                    "tracker_kwargs",
                    "results_folder",
                    "convert_image_to",
                    "save_and_sample_every",
                    "folder",
                    "tracker",
                ]
            ]
            parameters["dataset_name"] = Path(folder).name if folder else dataset.diagnosis

            # tracker_class = get_tracker_class(tracker.lower()) #TODO add tracker selection depending on the tracker name
            self.tracker = WandbTracker(
                project_name=tracker_kwargs.get("project_name", "medicraft"),
                hyperparameters=parameters,
                tags=tracker_kwargs.get("tags", None),
                # tags=getattr(tracker_kwargs, "tags", None),
                group=tracker_kwargs.get("group", "diffusion"),
                resume=tracker_kwargs.get("resume", None),
                id=tracker_kwargs.get("id", None),
                mode=tracker_kwargs.get("mode", "online"),
            )

        self.break_every_steps = None

        Path(results_folder).mkdir(parents=True, exist_ok=True)
        super().__init__(
            diffusion_model=diffusion_model,
            folder=folder,
            train_batch_size=train_batch_size,
            gradient_accumulate_every=gradient_accumulate_every,
            augment_horizontal_flip=augment_horizontal_flip,
            train_lr=train_lr,
            train_num_steps=train_num_steps,
            ema_update_every=ema_update_every,
            ema_decay=ema_decay,
            adam_betas=adam_betas,
            save_and_sample_every=save_and_sample_every,
            num_samples=num_samples,
            results_folder=results_folder,
            amp=amp,
            mixed_precision_type=mixed_precision_type,
            split_batches=split_batches,
            convert_image_to=convert_image_to,
            calculate_fid=calculate_fid,
            inception_block_idx=inception_block_idx,
            max_grad_norm=max_grad_norm,
            num_fid_samples=num_fid_samples,
            save_best_and_latest_only=save_best_and_latest_only,
        )
        if dataset is None:
            self.ds = _Dataset(
                folder,
                self.image_size,
                augment_horizontal_flip=augment_horizontal_flip,
                convert_image_to=convert_image_to,
            )
        else:
            self.ds = dataset
        print("Dataset initialized")
        assert (
            len(self.ds) >= 100
        ), "you should have at least 100 images in your folder. at least 10k images recommended"

        dl = DataLoader(self.ds, batch_size=train_batch_size, shuffle=True, pin_memory=True, num_workers=cpu_count())

        dl = self.accelerator.prepare(dl)
        self.dl = cycle(dl)
        print("Trainer initialized")

    def save(self, milestone: str, keep_last_models: int | None = 10):
        """
        Save the current state of the trainer.

        Args:
            milestone (str): The milestone identifier for the saved model.
            keep_last_models (int | None, optional): The number of last models to keep. Defaults to 10.
        """

        if not self.accelerator.is_local_main_process:
            return

        data = {
            "step": self.step,
            "model": self.accelerator.get_state_dict(self.model),
            "opt": self.opt.state_dict(),
            "ema": self.ema.state_dict(),
            "scaler": self.accelerator.scaler.state_dict() if exists(self.accelerator.scaler) else None,
            "version": __version__,
        }
        model_path = str(self.results_folder / f"model-{milestone}.pt")
        torch.save(data, model_path)
        torch.save(data, str(self.results_folder / "latest.pt"))
        self.keep_last_models(keep_last_models)

    def keep_last_models(self, num_models: int = 10) -> None:
        """
        Keep only the last `num_models` models in the results folder.
        Function will keep the latest model also.

        :param num_models: The number of models to keep.
        :type num_models: int
        """
        list_of_models = sorted(self.results_folder.glob("*.pt"), key=os.path.getmtime)
        models_to_remove = list_of_models[:-num_models]
        for model in models_to_remove:
            if model.stem == "latest":
                continue
            model.unlink()

    def load(self, milestone_path: str):
        """
        Load the model, optimizer, and other training state from a milestone file.

        :param milestone_path: The path to the milestone file.
        """

        accelerator = self.accelerator
        device = accelerator.device

        data = torch.load(milestone_path, map_location=device)

        model = self.accelerator.unwrap_model(self.model)
        model.load_state_dict(data["model"])

        self.step = data["step"]
        self.opt.load_state_dict(data["opt"])
        if self.accelerator.is_main_process:
            self.ema.load_state_dict(data["ema"])

        if "version" in data:
            print(f"loading from version {data['version']}")

        if exists(self.accelerator.scaler) and exists(data["scaler"]):
            self.accelerator.scaler.load_state_dict(data["scaler"])

    def train(self):
        """
        Train the model.

        This method performs the training loop for the model. It iterates over the specified number of training steps,
        accumulates gradients, updates the model parameters, and saves intermediate results if necessary.

        Returns:
            None
        """

        accelerator = self.accelerator
        device = accelerator.device

        with tqdm(initial=self.step, total=self.train_num_steps, disable=not accelerator.is_main_process) as pbar:
            while self.step < self.train_num_steps:
                total_loss = 0.0

                for _ in range(self.gradient_accumulate_every):
                    data = next(self.dl).to(device)

                    with self.accelerator.autocast():
                        loss = self.model(data)
                        loss = loss / self.gradient_accumulate_every
                        total_loss += loss.item()

                    self.accelerator.backward(loss)

                pbar.set_description(f"loss: {total_loss:.4f}")
                if self.tracker:
                    self.tracker.log({"loss": total_loss})

                accelerator.wait_for_everyone()
                accelerator.clip_grad_norm_(self.model.parameters(), self.max_grad_norm)

                self.opt.step()
                self.opt.zero_grad()

                accelerator.wait_for_everyone()

                self.step += 1
                if accelerator.is_main_process:
                    self.ema.update()

                    if self.step != 0 and divisible_by(self.step, self.save_and_sample_every):
                        self.ema.ema_model.eval()

                        with torch.inference_mode():
                            milestone = self.step // self.save_and_sample_every
                            batches = num_to_groups(self.num_samples, self.batch_size)
                            all_images_list = list(map(lambda n: self.ema.ema_model.sample(batch_size=n), batches))

                        all_images = torch.cat(all_images_list, dim=0)

                        if self.tracker:
                            self.tracker.log_images(all_images)

                        utils.save_image(
                            all_images,
                            str(self.results_folder / f"sample-{milestone}.png"),
                            nrow=int(math.sqrt(self.num_samples)),
                        )

                        # whether to calculate fid

                        if self.calculate_fid:
                            fid_score = self.fid_scorer.fid_score()
                            accelerator.print(f"fid_score: {fid_score}")
                        if self.save_best_and_latest_only:
                            if self.best_fid > fid_score:
                                self.best_fid = fid_score
                                self.save("best")
                            self.save("latest")
                        else:
                            self.save(milestone)
                        if self.break_every_steps and self.step % self.break_every_steps == 0:
                            self.save(f"{milestone}-val")

                pbar.update(1)
                if self.tracker:
                    self.tracker.update_step()

        accelerator.print("training complete")



import json
import logging
import sys

import yaml
from pydantic import BaseModel, ValidationError

import config as cfg
from pipeline.blocks import ConfigBlocks


def j_print(data, *args, **kwargs):
    """
    Print the given data in JSON format if possible, otherwise print it as is.
    For development purposes only.

    Args:
        data: The data to be printed.
        *args: Additional positional arguments to be passed to the print function.
        **kwargs: Additional keyword arguments to be passed to the print function.
    """

    try:
        print(json.dumps(data, indent=4), *args, **kwargs)
    except Exception:
        print(data, *args, **kwargs)


def read_config_file(config_file: str) -> dict:
    """
    Parse the configuration file

    :param config_file: The path to the configuration file
    :type config_file: str
    :return: The parsed configuration data
    :rtype: dict
    :raises yaml.YAMLError: If there is an error parsing the configuration file
    """
    try:
        with open(config_file, "r") as f:
            data = yaml.safe_load(f)
            return data
    except yaml.YAMLError as e:
        logging.error(f"Error parsing the configuration file: {e}")
        raise e


def parse_config(config: dict) -> dict[str, BaseModel]:
    """
    Parse and validate the configuration

    :param config: The configuration dictionary to be parsed and validated
    :type config: dict
    :return: A dictionary containing the parsed and validated configuration blocks
    :rtype: dict[str, BaseModel]
    """
    results = {}

    unique_essential_blocks = [ConfigBlocks.output.name, ConfigBlocks.general.name]

    for block in ConfigBlocks:
        block_config = config.get(block.name.lower())

        if block.name.lower() == ConfigBlocks.experiment.name.lower():
            block_config = {**block_config, **get_experiment_configs(config)}

        # get general configs
        if block.name.lower() not in unique_essential_blocks:
            general_config = config.get(ConfigBlocks.general.name)
            if general_config is None:
                logging.error("General config not found")
                sys.exit("Parsing config failed")

            general_block_config = general_config.get(block.name.lower())

            block_config = {**general_block_config, **block_config} if general_block_config else block_config

        try:
            block_instance: BaseModel = block.value(**block_config)
            results[block.name.lower()] = block_instance

        except ValidationError as e:
            logging.error(f"Error while processing block: {block.name} ")
            for e in e.errors():
                logging.error(
                    f"Error type: \033[1m{e['type']}\033[0m \tfor \033[1m{e['loc']}\033[0m\t| Error message: {e['msg']}"
                )
            sys.exit("Parsing config failed")
        if block.name == ConfigBlocks.experiment.name and cfg.DEV_DEBUG:
            j_print(block_instance.model_dump())
            print("OK", block.name)

    return results


def get_experiment_configs(config: dict) -> dict:
    """
    Get training configuration defined in other blocks.

    :param config: The configuration dictionary.
    :type config: dict
    :return: The experiment configuration dictionary.
    :rtype: dict
    """
    output_config = config.get(ConfigBlocks.output.name.lower())
    general_config = config.get(ConfigBlocks.general.name.lower())
    experiment_config = config.get(ConfigBlocks.experiment.name.lower())

    if output_config is None:
        logging.error("Output config not found")
        sys.exit("Parsing config failed")

    if general_config is None:
        logging.error("General config not found")
        sys.exit("Parsing config failed")

    if experiment_config is None:
        logging.error("Experiment config not found")
        sys.exit("Parsing config failed")

    experiment_config = {
        "total_steps": general_config.get("total_steps", 0),
        "image_size": general_config.get("image_size"),
        "experiment_id": general_config.get("experiment_id"),
        "models": general_config.get("models"),
        "results_dir": output_config.get("results_dir"),
        "copy_results_to": output_config.get("copy_results_to"),
        **experiment_config,
    }

    return experiment_config
